#!/usr/bin/env python

import numpy as np
import wplotlib 
import wuml
from sklearn.manifold import TSNE
from sklearn.decomposition import PCA
from sklearn.decomposition import KernelPCA
from sklearn.manifold import Isomap
from sklearn.manifold import LocallyLinearEmbedding
from sklearn.manifold import MDS
from sklearn.manifold import SpectralEmbedding
from sklearn.decomposition import FactorAnalysis

from wuml.type_check import *

class dimension_reduction:
	def __init__(self, data, n_components=2, method='PCA', learning_rate=30, show_plot=False, kernel='rbf', n_neighbors=5, gamma=1 ):
		'''
			n_components: number of dimension to reduce down to
			method: 'PCA', 'TSNE', 'KPCA', 'isoMap', 'LLE', 'MDS', 'Spectral Embedding','Factor Analysis'
			learning_rate: used for TSNE, if too large, everything become equidistance, defulat=30
			n_neighbors: used for isoMap
			gamma: used for KPCA
		'''

		df = ensure_DataFrame(data)
		X = ensure_numpy(data)
		if method == 'PCA':
			model = PCA(n_components=X.shape[1])
			use_all_dims = model.fit_transform(X)
			self.Ӽ = use_all_dims[:,0:n_components]

			self.normalized_eigs = model.explained_variance_ratio_
			self.eigen_values = model.singular_values_
			self.cumulative_eigs = np.cumsum(self.eigen_values)/np.sum(self.eigen_values)
			column_names = ['λ%d'%x for x in range(1, model.components_.shape[1]+1)]
			self.eig_vectors = ensure_DataFrame(model.components_, columns=column_names, index=df.columns)

		elif method == 'TSNE':
			model = TSNE(n_components=n_components, learning_rate=learning_rate)
			self.Ӽ = model.fit_transform(X)
		elif method == 'KPCA':
			model = KernelPCA(n_components=n_components, kernel=kernel, gamma=gamma)
			self.Ӽ = model.fit_transform(X)
		elif method == 'isoMap':
			model = Isomap(n_components=n_components, n_neighbors=n_neighbors)
			self.Ӽ = model.fit_transform(X)
		elif method == 'LLE':
			model = LocallyLinearEmbedding(n_components=n_components)
			self.Ӽ = model.fit_transform(X)
		elif method == 'MDS':
			model = MDS(n_components=n_components)
			self.Ӽ = model.fit_transform(X)
		elif method == 'Spectral Embedding':
			model = SpectralEmbedding(n_components=n_components)
			self.Ӽ = model.fit_transform(X)
		elif method == 'Factor Analysis':
			model = FactorAnalysis(n_components=n_components, random_state=0)
			self.Ӽ = model.fit_transform(X)
		else:
			raise ValueError('Error: Unrecognized Dimension Reduction Method: %s.'%method)

		self.model = model
		self.method = method
		self.shape = self.Ӽ.shape

		if show_plot:
			lp = wplotlib.scatter(self.Ӽ[:,0], self.Ӽ[:,1], 'Data After ' + method, 'X axis', 'Y axis', figsize=(10,5))		# (width, height)

	def __call__(self, X):
		methods = ['PCA', 'KPCA', 'isoMap','LLE', 'Factor Analysis']
		if self.method in methods:
			return self.model.transform(X)
		else:
			raise ValueError('This function does not have a transform function yet.')
		

	def __getitem__(self, item):
		return self.Ӽ[item]

	def __str__(self):
		return str(self.Ӽ)


def show_multiple_dimension_reduction_results(data, learning_rate=15, n_neighbors=5, gamma=1):
	methods = ['PCA', 'TSNE', 'KPCA', 'isoMap', 'LLE', 'MDS', 'Spectral Embedding','Factor Analysis']

	results = {}
	#lp = wplotlib.scatter(figsize=(7,8))		# (width, height)
	for i, m in enumerate(methods):
		results[m] = dimension_reduction(data, n_components=2, method=m, 
											learning_rate=learning_rate, show_plot=False, 
											n_neighbors=n_neighbors, gamma=gamma)

		if i == 0:
			lp = wplotlib.scatter(results[m].Ӽ[:,0], results[m].Ӽ[:,1], m, '', '', subplot=421 + i, show=False, figsize=(7,12))
		else:
			wplotlib.scatter(results[m].Ӽ[:,0], results[m].Ӽ[:,1], m, '', '', subplot=421 + i)

	lp.show()
	return results




import wuml
import numpy as np
import matplotlib.pyplot as plt

from sklearn.datasets import load_digits
from sklearn.neighbors import KernelDensity
from sklearn.decomposition import PCA
from sklearn.model_selection import GridSearchCV

from wuml.type_check import *


class KDE:
	def __init__(self, data):
		data = ensure_numpy(data)

		# use grid search cross-validation to optimize the bandwidth
		params = {'bandwidth': np.logspace(-1, 1, 20)}
		self.grid = GridSearchCV(KernelDensity(), params)
	
		self.grid.fit(data)
		self.σ = self.grid.best_estimator_.bandwidth
		self.kde = self.grid.best_estimator_

	def generate_samples(self, num_of_samples, return_data_type='wData'):
		samples = self.kde.sample(num_of_samples, random_state=0)
		return ensure_data_type(samples, type_name=return_data_type)

	def integrate(self, x0, x1):	# This only works for 1D data
		[result, error] = wuml.integrate(self, x0, x1)
		return result

	def __call__(self, data, return_log_likelihood=False):
		X = ensure_numpy(data)

		if return_log_likelihood: return self.kde.score_samples(X)
		else: return np.exp(self.kde.score_samples(X))




#!/usr/bin/env python

import numpy as np
import wuml


def L2X_data_A(N, xpath=None, ypath=None, Block_path=None):
	'''
		Warning: This data is an example of how L2X won't work b/c
					the underlying distribution between data of different
					feature contribution is the same.

		This function generates 4 dimensions {X1,X2,X3,X4} and 100 samples.
		The the y from 1st 50 samples only came from X1,X2
		The the y from 2nd 50 samples only came from X3,X4
		Standard regression cannot zero out half of the samples
	'''
	def gen_sample():
		X1 = np.random.uniform(0,2) + 5
		X2 = np.random.uniform(0,2) + 5
		X3 = np.random.exponential(scale=1.0)
		X4 = np.random.randn()
	
		return np.array([[X1,X2,X3,X4]])

	b = np.array([[2],[1],[3],[3]])
	X = np.empty((0,4))
	for n in range(N):
		X = np.vstack((X,gen_sample()))

	
	O = np.ones((int(N/2),2))
	Z = np.zeros((int(N/2),2))
	B = np.block([[O, Z],[Z,O]])
	y = (X*B).dot(b)

	if xpath is not None: np.savetxt(xpath, X, delimiter=',', fmt='%.4f') 
	if ypath is not None: np.savetxt(ypath, y, delimiter=',', fmt='%.4f') 
	if Block_path is not None: np.savetxt(Block_path, B, delimiter=',', fmt='%d') 

	return [X,y,b]
	
def L2X_data_B(N, xpath=None, ypath=None, Block_path=None):
	'''
		Note: This data is an example of how L2X will work b/c
					the underlying distribution between data of different
					feature contribution are different.

		This function generates 4 dimensions {X1,X2,X3,X4} and 100 samples.
		The the y from 1st 50 samples only came from X1,X2
		The the y from 2nd 50 samples only came from X3,X4
		Standard regression cannot zero out half of the samples
	'''
	def gen_samples_A():
		X1 = np.random.beta(0.5,0.5) + 10
		X2 = np.random.beta(0.5,0.5) + 10
		X3 = np.random.beta(0.5,0.5) + 10
		X4 = np.random.rayleigh() + 10
	
		return np.array([[X1,X2,X3,X4]])


	def gen_samples_B():
		X1 = 3*np.random.uniform(0,2) 
		X2 = 3*np.random.uniform(0,2) 
		X3 = 3*np.random.exponential(scale=1.0)
		X4 = 3*np.random.randn()
	
		return np.array([[X1,X2,X3,X4]])

	Ns = int(N/2)

	b = np.array([[2],[1],[3],[3]])
	X = np.empty((0,4))
	for n in range(int(Ns)):
		X = np.vstack((X,gen_samples_A()))
	for n in range(int(Ns)):
		X = np.vstack((X,gen_samples_B()))

	O = np.ones((int(Ns),2))
	Z = np.zeros((int(Ns),2))
	B = np.block([[O, Z],[Z,O]])
	y = (X*B).dot(b)

	if xpath is not None: np.savetxt(xpath, X, delimiter=',', fmt='%.4f') 
	if ypath is not None: np.savetxt(ypath, y, delimiter=',', fmt='%.4f') 
	if Block_path is not None: np.savetxt(Block_path, B, delimiter=',', fmt='%d') 

	return [X,y,b]

X = L2X_data_B(5000, xpath='l2x_synthetic.csv', ypath='l2x_synthetic_label.csv', Block_path='l2x_synthetic_blocks.csv')



#!/usr/bin/env python

import torch
import numpy as np
import sys
from wuml.IO import *
from sklearn import linear_model
from torch import nn
from torch.autograd import Variable
from inspect import signature
import torch.nn.functional as F
import torch.nn as nn
import collections

class flexable_Model(torch.nn.Module):
	def __init__(self, dataDim, networkStructure):
		super(flexable_Model, self).__init__()

		self.networkStructure = networkStructure
		inDim = dataDim
		for l, Layer_info in enumerate(networkStructure):
			layer_width, activation_function = Layer_info
			φ = activation_function
			outDim = layer_width

			lr = 'self.l' + str(l) + ' = torch.nn.Linear(' + str(inDim) + ', ' + str(outDim) + ' , bias=True)'
			exec(lr)
			exec('self.l' + str(l) + '.activation = "' + φ + '"')		#softmax, relu, tanh, sigmoid, none
			inDim = outDim

	def forward(self, x):
		self.y0 = x
		for m, layer in enumerate(self.children(),0):
			var = 'self.y' + str(m+1)

			if layer.activation == 'none':
				cmd = 'self.yout = ' + var + ' = self.l' + str(m) + '(self.y' + str(m) + ')'
			else:
				cmd = 'self.yout = ' + var + ' = F.' + layer.activation + '(self.l' + str(m) + '(self.y' + str(m) + '))'
			exec(cmd)

		if torch.isnan(self.yout).any():
			print('\n nan was detected inside a network forward\n')
			import pdb; pdb.set_trace()

		return self.yout


def run_SGD(loss_function, model_parameters, trainLoader, device, 
				X_dataType=torch.FloatTensor, Y_dataType=torch.FloatTensor,
				model=None, lr=0.001, print_status=True, max_epoch=1000,
				on_new_epoch_call_back=None):

	optimizer = torch.optim.Adam(model_parameters, lr=lr)	
	scheduler = torch.optim.lr_scheduler.ReduceLROnPlateau( optimizer, factor=0.5, min_lr=1e-10, patience=50, verbose=False)

	# For all loss check https://neptune.ai/blog/pytorch-loss-functions
	if type(loss_function).__name__ == 'str':
		if loss_function == 'mse':
			loss_function = nn.MSELoss()
		elif loss_function == 'L1':
			loss_function = nn.L1Loss()
		elif loss_function == 'CE':
			loss_function = nn.CrossEntropyLoss()
		elif loss_function == 'hindge':
			loss_function = nn.HingeEmbeddingLoss()

	paramLen = len(signature(loss_function).parameters) # number of arguments


	for epoch in range(max_epoch):

		loss_list = []	
		for (i, data) in enumerate(trainLoader):
			[x, y, ind] = data

			x = Variable(x.type(X_dataType), requires_grad=False)
			y = Variable(y.type(Y_dataType), requires_grad=False)
			x= x.to(device, non_blocking=True )
			y= y.to(device, non_blocking=True )
			optimizer.zero_grad()

			if model is not None:
				ŷ = model(x)
				ŷ = torch.squeeze(ŷ)
				y = torch.squeeze(y)

				if paramLen == 4:
					loss = loss_function(x, y, ŷ, ind)
				elif paramLen == 3:
					loss = loss_function(x, y, ind)
				elif paramLen == 2:
					#loss = loss_function(y, ŷ)
					loss = loss_function(ŷ,y)
			else:
				try:
					loss = loss_function(x, y, ind)
				except:
					loss = loss_function(x, y)

			if torch.isnan(loss): import pdb; pdb.set_trace()

			loss.backward()
			optimizer.step()

			loss_list.append(loss.item())

		loss_avg = np.array(loss_list).mean()
		scheduler.step(loss_avg)
		if print_status:
			txt = '\tepoch: %d, Avg Loss: %.4f, Learning Rate: %.8f'%((epoch+1), loss_avg, scheduler._last_lr[0])
			write_to_current_line(txt)

		if on_new_epoch_call_back is not None:
			on_new_epoch_call_back(loss_avg, (epoch+1), scheduler._last_lr[0])

		if model is not None:
			if "on_new_epoch" in dir(model):
				early_exit = model.on_new_epoch(loss_avg, (epoch+1), scheduler._last_lr[0])
				if early_exit: break



class basicNetwork:
	def __init__(self, costFunction, X, 
						Y=None, networkStructure=[(3,'relu'),(3,'relu'),(3,'none')], 
						on_new_epoch_call_back = None, max_epoch=1000, 	X_dataType=torch.FloatTensor, 
						Y_dataType=torch.FloatTensor, learning_rate=0.001, simplify_network_for_storage=None,
						network_usage_output_type='Tensor', network_usage_output_dim='none', network_info_print=False): 
		'''
			X : This should be wData type
			possible activation functions: softmax, relu, tanh, sigmoid, none
			simplify_network_for_storage: if a network is passed as this argument, we create a new network strip of unnecessary stuff
			network_usage_output_dim: network output dimension, 0, 1 or 2
		'''
		self.network_usage_output_type = network_usage_output_type
		self.network_usage_output_dim = network_usage_output_dim

		if simplify_network_for_storage is None:
			#	X should be in wuml format
			self.trainLoader = X.get_data_as('DataLoader')

			self.lr = learning_rate
			self.max_epoch = max_epoch
			self.X_dataType = X_dataType
			self.Y_dataType = Y_dataType
			self.costFunction = costFunction
			self.NetStructure = networkStructure
			self.on_new_epoch_call_back = on_new_epoch_call_back #set this as a callback at each function
			self.model = flexable_Model(X.shape[1], networkStructure)
			self.network_output_in_CPU_during_usage = False
		else:
			self.costFunction = costFunction
			self.on_new_epoch_call_back = on_new_epoch_call_back #set this as a callback at each function

			θ = simplify_network_for_storage
			self.lr = θ.lr
			self.max_epoch = θ.max_epoch
			self.X_dataType = θ.X_dataType
			self.Y_dataType = θ.Y_dataType
			self.NetStructure = θ.NetStructure
			self.model = θ.model
			self.network_output_in_CPU_during_usage = True

		if X.label_type == 'discrete': self.Y_dataType = torch.int64		#overide datatype if discrete labels

		if torch.cuda.is_available(): 
			self.device = 'cuda'
			self.model.to(self.device)		# store the network weights in gpu or cpu device
		else: self.device = 'cpu'

		self.out_structural = None
		self.info(printOut=network_info_print)
		self.network_info_print = network_info_print

		#	Catch some errors
		if costFunction == 'CE' and X.label_type == 'continuous':
			print("\n the data label_type should not be continuous when using 'CE' as costFunction during classification!!!\n")
		elif costFunction == 'hindge' and X.label_type == 'continuous':
			print("\n the data label_type should not be continuous when using 'hindge' as costFunction during classification!!!\n")

	def info(self, printOut=True):
		 
		info_str ='Network Info:\n'
		info_str += '\tLearning rate: %.3f\n'%self.lr
		info_str += '\tMax number of epochs: %d\n'%self.max_epoch
		info_str += '\tCost Function: %s\n'%str(self.costFunction)
		info_str += '\tTrain Loop Callback: %s\n'%str(self.on_new_epoch_call_back)
		info_str += '\tCuda Available: %r\n'%torch.cuda.is_available()
		info_str += '\tNetwork Structure\n'
		for i in self.model.children():
			try:
				info_str += ('\t\t%s , %s\n'%(i,i.activation))
			except:
				info_str += ('\t\t%s \n'%(i))
		if printOut: print(info_str)
		return info_str

	def predict(self, data):
		return self.__call__(data, output_type='ndarray')

	def __call__(self, data, output_type='Tensor', out_structural=None):
		'''
			out_structural (mostly for classification purpose): None, '1d_labels', 'one_hot'
		'''

		if out_structural is None: out_structural = self.out_structural
		if type(data).__name__ == 'ndarray': 
			x = torch.from_numpy(data)
			x = Variable(x.type(self.X_dataType), requires_grad=False)
			x= x.to(self.device, non_blocking=True )
		elif type(data).__name__ == 'Tensor': 
			x = Variable(x.type(self.X_dataType), requires_grad=False)
			x= x.to(self.device, non_blocking=True )
		elif type(data).__name__ == 'wData': 
			x = data.get_data_as('Tensor')
		else:
			raise
		yout = self.model(x)

		if self.network_usage_output_dim == 0 or self.network_usage_output_dim == 1:
			yout = torch.squeeze(yout)
		if self.network_usage_output_dim == 2:
			yout = torch.atleast_2d(yout)


		if out_structural == '1d_labels':
			_, yout = torch.max(yout, 1)
		elif out_structural == 'one_hot':
			_, yout = torch.max(yout, 1)
			yout = wuml.one_hot_encoding(yout, output_data_type='same')
		elif out_structural == 'softmax':
			m = nn.Softmax(dim=1)
			yout = m(yout)

		if output_type == 'ndarray' or self.network_usage_output_type == 'ndarray':
			return yout.detach().cpu().numpy()
		elif self.network_output_in_CPU_during_usage:
			return yout.detach().cpu()

		return yout

	def eval(self, output_type='ndarray', out_structural=None):		#	Turn this on to run test results
		self.network_usage_output_type = output_type
		if out_structural is not None: self.out_structural = out_structural
		self.model.eval()

	def train(self, print_status=True):
		param = self.model.parameters()
		[ℓ, TL, mE, Dev] = [self.costFunction, self.trainLoader, self.max_epoch, self.device]
		[Xtype, Ytype] = [self.X_dataType, self.Y_dataType]

		run_SGD(ℓ, param, TL, Dev, model=self.model, lr=self.lr, print_status=print_status,
				max_epoch=mE, X_dataType=Xtype, Y_dataType=Ytype, 
				on_new_epoch_call_back = self.on_new_epoch_call_back)

	def fit(self, X,Y):
		self.train(print_status=self.network_info_print)





import sklearn.metrics
from wuml.type_check import *


class IKDR:
	'''
	Interpretable Kernel Dimension Reduction
	'''

	def __init__(self, data, q=2, y=None, y_column_name=None, kernel='rbf', max_rep=8, rank_constrain_ratio=0):
		'''
			rank_constrain_ratio: this constraint lowers the rank of W, default 0 removes the constraint
			q: the number of dimension you want to reduce down to
		'''
		self.X = X = ensure_numpy(data)
		self.Y = Y = ensure_label(data, y=y, y_column_name=y_column_name)
		self.σ = np.median(sklearn.metrics.pairwise.pairwise_distances(X))

		Yₒ = wuml.one_hot_encoding(Y)
		Kᵧ = Yₒ.dot(Yₒ.T)
		self.Γ = wuml.double_center(Kᵧ)	# HKᵧH
		self.ζ = rank_constrain_ratio
		self.max_rep = max_rep
		self.conv_threshold = 0.01
		self.q = q

		[self.W, self.λ] = self.update_W(q, W=None)

	def update_W(self, q=None, W=None):
		X, Γ, σ = self.X, self.Γ, self.σ

		Φ = self.gaussian_Φ(X, Γ, σ, W=W)
		return self.eig_solver(Φ, q)


	def eig_solver(self, L, q, mode='smallest'):
		eigenValues,eigenVectors = np.linalg.eigh(L)
	
		if mode == 'smallest':
			U = eigenVectors[:, 0:q]
			U_λ = eigenValues[0:q]
		elif mode == 'largest':
			n2 = len(eigenValues)
			n1 = n2 - q
			U = eigenVectors[:, n1:n2]
			U_λ = eigenValues[n1:n2]
		else:
			raise ValueError('unrecognized mode : ' + str(mode) + ' found.')
		
		return [U, U_λ]



	def gaussian_Φ(self, X, Γ, σ, W=None):
		Ɗ = wuml.compute_Degree_matrix(Γ)
	
		if W is None:
			Φ = X.T.dot(Ɗ - Γ).dot(X)
		else:	
			Kx = wuml.rbk_kernel(X.dot(W), σ)
			Ψ=Γ*Kx
			D_Ψ = wuml.compute_Degree_matrix(Ψ)
			Φ = X.T.dot(D_Ψ - Ψ).dot(X) 			#debug.compare_Φ(db, Φ, Ψ)
			Φ = self.add_rank_constraint(Φ)

		return Φ

	def add_rank_constraint(self, Φ):		# from the log det rank constraint
		if self.ζ == 0: return Φ
		W, X, ζ = self.W, self.X, self.ζ
	
		A = W.dot(W.T) + 0.001*np.eye(X.shape[1])
		rc = np.linalg.inv(A)
	
		Φ_norm = np.linalg.norm(Φ)
		rc_norm = np.linalg.norm(rc)
		return Φ + (ζ*Φ_norm/rc_norm)*rc

	def fit(self, data, Y):
		Γ, σ, q, λᑊ = self.Γ, self.σ, self.q, self.λ
		X = ensure_numpy(data)

		for i in range(self.max_rep):
			[Ŵ, λᒾ] = self.update_W(q=q, W=self.W)
			Δ = np.linalg.norm(λᑊ- λᒾ)/np.linalg.norm(λᑊ)
			H = wuml.HSIC(X.dot(Ŵ),Y)

			self.W = Ŵ
			λᑊ = λᒾ

			if Δ < self.conv_threshold: break

		reduced_X = X.dot(self.W)
		self.classifier = wuml.classification(reduced_X, y=Y, classifier='SVM')

	def predict(self, data):
		X = ensure_numpy(data)
		reduced_X = X.dot(self.W)

		return self.classifier(reduced_X)




import wuml
import numpy as np
import matplotlib.pyplot as plt
import torch
from torch import nn
from wuml.type_check import *




class RealNVP(nn.Module):
	def __init__(self, nets, nett, num_flows, prior, D=2, dequantization=True):
		super(RealNVP, self).__init__()
		self.dequantization = dequantization
		
		self.prior = prior
		self.t = torch.nn.ModuleList([nett() for _ in range(num_flows)])
		self.s = torch.nn.ModuleList([nets() for _ in range(num_flows)])
		self.num_flows = num_flows
		
		self.D = D

	def coupling(self, x, index, forward=True):
		# x: input, either images (for the first transformation) or outputs from the previous transformation
		# index: it determines the index of the transformation
		# forward: whether it is a pass from x to y (forward=True), or from y to x (forward=False)
		
		(xa, xb) = torch.chunk(x, 2, 1)
		
		s = self.s[index](xa)
		t = self.t[index](xa)
		
		if forward:
			#yb = f^{-1}(x)
			yb = (xb - t) * torch.exp(-s)
		else:
			#xb = f(y)
			yb = torch.exp(s) * xb + t
		
		return torch.cat((xa, yb), 1), s

	def permute(self, x):
		return x.flip(1)

	def f(self, x):
		log_det_J, z = x.new_zeros(x.shape[0]), x
		for i in range(self.num_flows):
			z, s = self.coupling(z, i, forward=True)
			z = self.permute(z)
			log_det_J = log_det_J - s.sum(dim=1)

		return z, log_det_J

	def f_inv(self, z):
		x = z
		for i in reversed(range(self.num_flows)):
			x = self.permute(x)
			x, _ = self.coupling(x, i, forward=False)

		return x

	def forward(self, x, reduction='avg'):
		z, log_det_J = self.f(x)
		if reduction == 'sum':
			return -(self.prior.log_prob(z) + log_det_J).sum()
		else:
			return -(self.prior.log_prob(z) + log_det_J).mean()

	def sample(self, batchSize):
		z = self.prior.sample((batchSize, self.D))
		z = z[:, 0, :]
		x = self.f_inv(z)
		return x.view(-1, self.D)


class flow:
	def __init__(self, data=None, flow_mothod='realNVP', network_width=512, output_model_name='flow', print_status=True, 
					lr = 1e-3, max_epochs=1000, max_patience=None, num_flows=8, dequantization=False, 
					load_model_path=None):
		'''
			max_patience: an early stopping is used, if training doesn't improve for longer than 20 epochs, it is stopped
		'''

		self.netW = M = network_width
		self.lr = lr
		self.max_epochs = max_epochs
		self.max_patience = max_patience
		self.print_status = print_status

		# get data
		self.data = data
		X = ensure_wData(data)

		if data.Y is None: data.Y = data.X

		[X_train, X_test, y_train, y_test] = wuml.split_training_test(X, test_percentage=0.2, save_as='none')
		X_train_DL = X_train.get_data_as('DataLoader')
		X_test_DL = X_test.get_data_as('DataLoader')
		d = X.shape[1]

		# scale (s) network
		self.S = nets = lambda: nn.Sequential(nn.Linear(d // 2, M), nn.LeakyReLU(), 
										nn.Linear(M, M), nn.LeakyReLU(), 
										nn.Linear(M, M), nn.LeakyReLU(), 
										nn.Linear(M, M), nn.LeakyReLU(), 
										nn.Linear(M, M), nn.LeakyReLU(), 
										nn.Linear(M, d // 2), nn.Tanh())
		
		# translation (t) network
		self.T = nett = lambda: nn.Sequential(nn.Linear(d // 2, M), nn.LeakyReLU(), 
										nn.Linear(M, M), nn.LeakyReLU(), 
										nn.Linear(M, M), nn.LeakyReLU(), 
										nn.Linear(M, M), nn.LeakyReLU(), 
										nn.Linear(M, M), nn.LeakyReLU(), 
										nn.Linear(M, d // 2))

		if torch.cuda.is_available(): self.device = 'cuda'
		else: self.device = 'cpu'

		# Prior (a.k.a. the base distribution): Gaussian
		μ = torch.zeros(d)
		σ = torch.eye(d)
		μ = μ.to(self.device, non_blocking=True )
		σ = σ.to(self.device, non_blocking=True )
		self.prior = prior = torch.distributions.MultivariateNormal(μ, σ)

		# Init RealNVP
		if load_model_path is not None:
			self.model = torch.load(load_model_path)
			#self.model = wuml.pickle_load(load_model_path)
			self.model.eval()
		else:
			self.model = model = RealNVP(nets, nett, num_flows, prior, D=d, dequantization=dequantization)
			self.model.to(self.device)		# store the network weights in gpu or cpu device

			# OPTIMIZER
			self.optimizer = optimizer = torch.optim.Adamax([p for p in model.parameters() if p.requires_grad == True], lr=lr)

			# Training procedure
			nll_val = self.train(name=output_model_name, max_patience=max_patience, num_epochs=max_epochs, 
								model=model, optimizer=optimizer, training_loader=X_train_DL, val_loader=X_test_DL)

	def save(self, pth): 
		torch.save(self.model, pth)

	def train(self, name, max_patience, num_epochs, model, optimizer, training_loader, val_loader):
		nll_val = []
		best_nll = 1000.
		patience = 0

		X_dataType = self.data.xtorchDataType
		Y_dataType = self.data.ytorchDataType

		# Main loop
		for e in range(num_epochs):
			# TRAINING
			model.train()
			for indx_batch, data in enumerate(training_loader):
				[x, y, ind] = data
	
				x = Variable(x.type(X_dataType), requires_grad=False)
				y = Variable(y.type(Y_dataType), requires_grad=False)
				x= x.to(self.device, non_blocking=True )
				y= y.to(self.device, non_blocking=True )

				if hasattr(model, 'dequantization'):
					if model.dequantization:
						x = x + (1. - torch.rand(x.shape))/2.
				loss = model.forward(x)
	
				optimizer.zero_grad()
				loss.backward(retain_graph=True)
				optimizer.step()
	
			# Validation
			loss_val = self.evaluation(val_loader, model_best=model, epoch=e)
			nll_val.append(loss_val)  # save for plotting
	
			#	Print training status	
			if self.print_status:
				txt = '\tepoch: %d, Avg Loss: %.4f, patience: %d'%((e), loss, patience)
				wuml.write_to_current_line(txt)

			if e == 0:
				#torch.save(model, name + '.model')
				best_nll = loss_val
			else:
				if loss_val < best_nll:
					#torch.save(model, name + '.model')
					best_nll = loss_val
					patience = 0
	
					#samples_generated(name, val_loader, extra_name="_epoch_" + str(e))
				else:
					patience = patience + 1
	
			if max_patience is not None:
				if patience > max_patience:
					print('\n\tFinal epoch: %d, Avg Loss: %.4f, patience: %d'%((e), loss, patience))
					break
	
		nll_val = np.asarray(nll_val)
	
		return nll_val

	def evaluation(self, test_loader, name=None, model_best=None, epoch=None):
		# EVALUATION
		if model_best is None:
			# load best performing model
			model_best = torch.load(name + '.model')

		X_dataType = self.data.xtorchDataType
		Y_dataType = self.data.ytorchDataType

		model_best.eval()
		loss = 0.
		N = 0.
		for indx_batch, test_batch in enumerate(test_loader):
			[x, y, ind] = test_batch

			x = Variable(x.type(X_dataType), requires_grad=False)
			y = Variable(y.type(Y_dataType), requires_grad=False)
			x= x.to(self.device, non_blocking=True )
			y= y.to(self.device, non_blocking=True )


			if hasattr(self.model, 'dequantization'):
				if self.model.dequantization:
					x = x + (1. - torch.rand(x.shape))/2.
			loss_t = model_best.forward(x, reduction='sum')
			loss = loss + loss_t.item()
			N = N + x.shape[0]
		loss = loss / N
	
		return loss

	def generate_samples(self, num_of_samples, return_data_type='wData'):
		samples = self.model.sample(num_of_samples)
		return ensure_data_type(samples, type_name=return_data_type)


	#def integrate(self, x0, x1):	# This only works for 1D data
	#	[result, error] = wuml.integrate(self, x0, x1)
	#	return result

	def __call__(self, data, return_log_likelihood=False, return_data_type='wData'):
		x = ensure_tensor(data)

		z, log_det_J = self.model.f(x)
		logLike = self.prior.log_prob(z) + log_det_J

		if return_log_likelihood: output = logLike
		else: output = torch.exp(logLike)
		
		return ensure_data_type(output, type_name=return_data_type)

